(ns lu-preprocessing.map
 (:require [clojure.string :as string]))


;;; private defs

;; Defines regular expression to delimit values from file lines.
(def ^:private delimiter #"\s")


;;; private methods

(defn- parse-positions
  "Parses value positions from the map matrix.
      values: List of searched values.
      map-matrix: A map matrix.
      na-value: NA value.
   returns: A hash-map of values and their by-row positions."
  [[values
    map-matrix
    na-value]]
  (let [row-size
        (count (first map-matrix))]
    (->> (flatten map-matrix)
         (map-indexed vector)
         (reduce (fn[seed [index value]]
                   (if (= value na-value)
                     seed
                     (->> ((juxt quot
                                 mod)
                           index
                           row-size)
                          ((fn[[i j]]
                            (update-in seed
                                       [value
                                        i]
                                       #((fnil conj
                                               [])
                                         %
                                         j)))))))
                 {}))))


;;; public methods

(defn parse
  "Parses positions of cell values.
   Cells should contain only integer values.
      map-lines: A list of .asc file rows.
   returns: A hash-map of cell values and their positions.

   ## Usage

      (require '[lu-preprocessing.map :refer :all])

      (parse [\"ncols 5\"
              \"nrows 5\"
              \"xllcorner 1.0\"
              \"yllcorner 2.0\"
              \"cellsize 200.0\"
              \"NODATA_value -9999\"
              \"-9999 1 2 1 -9999\"
              \"0 0 0 0 0\"
              \"0 -9999 0 -9999 1\"
              \"1 1 -9999 2 -9999\"
              \"-9999 -9999 -9999 -9999 -9999\"])
      => {0 {1 [0 1 2 3 4]
             2 [0 2]}
          1 {0 [1 3]
             2 [4]
             3 [0 1]}
          2 {0 [2]
             3 [3]}}
  "
  [map-lines]
  (let [split #(string/split % delimiter)]
    (->> map-lines
         ((juxt #(->> (drop 5 %)
                      first
                      split
                      second)
                #(->> (drop 6 %)
                      (map split))))
         ((juxt #(->> (second %)
                      flatten
                      distinct
                      (remove (partial = (first %))))
                second
                first))
         parse-positions
         ((juxt (fn[positions]
                  (->> (keys positions)
                       (map #(Integer/parseInt %))))
                vals))
         (apply zipmap))))
