(ns lu-rescaling.map-operations
  (:require [clojure.math.combinatorics :as combo]))


;;; private methods

(defn- get-by-positions
  "Returns a collection of value lists
            located at the corresponding positions.
      maps: A list of hash-maps with location-value pairs.
      rows: A hash-map with row keys and corresponding columns.
   returns: A hash-map with location-values pairs."
  [maps rows]
  (->> (map (fn[[row columns]]
              (map vector
                   (repeat row)
                   columns))
            rows)
       (apply concat)
       (map (fn [row-column-pair]
              (->> (vals maps)
                   (map #(get % row-column-pair))
                   vec)))
       (apply map vector)
       (zipmap (keys maps))))

(defn- get-positions
  "Returns a list of positions.
    rows: A hash-map with row keys and corresponding columns.
   returns: A list of positions (row-column pairs)."
  [rows]
  (->> ((juxt #(map (fn[[v1 v2]]
                      (repeat (count v2)
                              v1))
                    %)
              vals)
        rows)
       (map #(apply concat %))
       (apply map vector)))


;;; public methods

(defn reverse-map
  "Returns a reversed map with position keys.
      source-map: A map with value keys.
   returns: A reversed hash-map with position keys.

  ## Usage

      (require '[lu-rescaling.map-operations :refer :all])

      (def map {1 {0 [0 1 2] 1 [1]}
                2 {1 [0] 2 [1 2]}
                3 {1 [2] 2 [0]}})

      (reverse-map map)
      => {[0 0] 1
          [0 1] 1
          [0 2] 1
          [1 0] 2
          [1 1] 1
          [1 2] 3
          [2 0] 3
          [2 2] 2}
  "
  [source-map]
  (->> source-map
       (map (fn[[value rows]]
              (->> (get-positions rows)
                   (map vector
                        (repeat value))
                   (map reverse)
                   (apply concat))))
       (apply concat)
       (apply hash-map)))

(defn merge-into-cross-map
  "Returns a cross map of values.
      source-map: A source hash-map, whose keys are identifiers.
      target-maps: A list of hash-maps with reference values.
   returns: A hash-map with values from the source map as keys
            and values from the other maps which are located
            at the same positions as values.

  ## Usage

      (require '[lu-rescaling.map-operations :refer :all])

      (def source-map {0 {0 [1 2] 1 [0 1 2] 2 [2]}
                       1 {0 [0] 2 [0 1]}})
      (def target-maps {\"id1\" {0 {2 [1]} 1 {1 [2]} 2 {1 [0 1] 2 [0]} 3 {0 [0 1 2] 2 [2]}}
                        \"id2\" {1 {0 [1 2] 1 [0 1 2]} 2 {0 [0] 2 [0 1 2]}}})

      (merge-into-cross-map source-map target-maps)
      => {0 {\"id1\" [3 3 2 2 1 3]
             \"id2\" [1 1 1 1 1 2]}
          1 {\"id1\" [3 2 0]
             \"id2\" [2 2 2]}}
  "
  [source-map target-maps]
  (let [maps
        (->> target-maps
             ((juxt keys
                    #(->> (vals %)
                          (map reverse-map))))
             (apply zipmap))]
    (->> (vals source-map)
         (map #(get-by-positions maps %))
         (zipmap (keys source-map)))))
