(ns lu-rescaling.reader
 (:require [clojure.string :as string]
           [clojure.java.io :as io]
           [utilities.reader :as reader]
           [utilities.file :as file]))


;;; private defs

;; Defines regular expression to delimit values from file lines.
(def ^:private delimiter #"\s")

;;; private methods

(defn- parse-positions
  "Parses value positions in the map matrix.
      values: List of searched values.
      map-matrix: A map matrix.
      na-value: NA value.
   returns: A hash-map of values and their by-row positions."
  [[values
    map-matrix
    na-value]]
  (let [row-size
        (count (first map-matrix))]
    (->> (flatten map-matrix)
         (map-indexed vector)
         (reduce (fn[seed [index value]]
                   (if (= value
                          na-value)
                     seed
                     (->> ((juxt quot
                                 mod)
                           index
                           row-size)
                          ((fn[[i j]]
                            (update-in seed
                                       [value
                                        i]
                                       #((fnil conj
                                               [])
                                         %
                                         j)))))))
                 {}))))

(defn- parse-map
  "Parses a map matrix from the .asc file.
      path: Path to the raster file.
      int?: A value indicating whether value type is integer.
   returns: A hash-map with values and list of
            their (row, column) positions."
  [path int?]
  (let [split #(string/split % delimiter)]
    (->> (reader/read-file path)
         ((juxt #(->> (drop 5 %)
                      first
                      split
                      second)
                #(->> (drop 6 %)
                      (map split))))
         ((juxt #(->> (second %)
                      flatten
                      distinct
                      (remove (partial = (first %))))
                second
                first))
         parse-positions
         ((juxt (fn[positions]
                  (->> (keys positions)
                       (map #(if int?
                               (Integer/parseInt %)
                               (Double/parseDouble %)))))
                vals))
         (apply zipmap))))


;;; public methods

(defn parse-maps
  "Reads content of .asc files from the folder.
      folder: Path to the source folder.
   returns: A hashmap with file names as keys and
            hashmaps of values and their row-column positions as values.

   ## Usage

        (require '[lu-rescaling.reader :refer :all])

        (parse-maps \"GIS raster maps\")
  "
  [folder]
  (->> (file/walk folder #".*\.asc")
       ((juxt (fn[files]
                (map #(.getName %)
                     files))
              (fn[files]
                (map #(parse-map % false)
                     files))))
       (apply zipmap)))

(defn parse-region-map
  "Reads content of .asc file with regional administrative division from the path.
      path: Path to the source .asc file.
   returns: A hashmap row-column positions per identifier of an administrative unit.

  ## Usage

        (require '[lu-rescaling.reader :refer :all])

        (parse-region-map \"sections.asc\")
  "
  [path]
  (let [id (.getName (io/file path))]
    (->> (parse-map path true)
         (hash-map id))))
