(ns data-cleaning.filter)


;;; private methods

(defn- mask-in-a-row
  "Masks cells in the row.
      map-vector: A vector of map cell values.
      number-of-columns: A number of columns in the map.
      row: A row number.
      columns: A list of column numbers in the rows.
   returns: Map with the selected masked cells in the row."
  [map-vector
   number-of-columns
   row
   columns]
  (reduce (fn [seed column]
            (assoc seed
                   (+ (* row number-of-columns)
                      column)
                   nil))
          map-vector
          columns))

(defn- get-neighbour-positions
  "Returns neighbours posiitions for the selected cell.
      row: A cell row number.
      column: A cell column number.
      max-row: A maximum row number in the map.
      max-column: A maximum column number in the map.
   returns: Neighbours row and column numbers."
  [row
   column
   max-row
   max-column]
  (let [get-levels
        (fn [value max-value]
          (condp = value
            0 [value (inc value)]
            max-value [(dec value) value]
            [(dec value) value (inc value)]))]
    {:rows (get-levels row
                       max-row)
     :columns (get-levels column
                          max-column)}))

(defn- mask-self-and-neighbours
  "Masks a cell and it's Moore neighbourhood.
      map-vector: A vector of map cell values.
      row: A row number.
      column: A column number.
      number-of-columns: A number of columns in the map.
   returns: Map with the masked cell and it's Moore neighbourhood."
  [map-vector
   row
   column
   number-of-columns]
  (let [neighbours
        (get-neighbour-positions row
                                 column
                                 (-> (count map-vector)
                                     (/ number-of-columns)
                                     dec)
                                 (dec number-of-columns))]
    (reduce #(mask-in-a-row %1
                            number-of-columns
                            %2
                            (:columns neighbours))
            map-vector
            (:rows neighbours))))

(defn- mask-na
  "Masks cells with na-values and their Moore neighbourhood.
      map-vector: A vector of map cell values.
      number-of-columns: A number of columns in the map.
      na-values: List of na values per map row.
   returns: Map with masked na cells and their Moore neighbourhood."
  [map-vector
   number-of-columns
   na-values]
  (if (empty? na-values)
    map-vector
    (recur (let [[row columns] (first na-values)]
             (reduce #(mask-self-and-neighbours %1
                                                row
                                                %2
                                                number-of-columns)
                     map-vector
                     columns))
           number-of-columns
           (next na-values))))

(defn- mask-row
  "Masks all cells in the selected row.
      map-vector: A vector of map cell values.
      row: A row number.
      number-of-columns: A number of columns in the map.
   returns: Map with the masked row."
  [map-vector
   row
   number-of-columns]
  (mask-in-a-row map-vector
                 number-of-columns
                 row
                 (range number-of-columns)))

(defn- mask-first-and-last-elements
  "Masks cells, which are the first and
   last elements in the selected map rows.
      map-vector: A vector of map cell values.
      rows: List of map rows.
      number-of-columns: A number of columns in the map.
   returns: Map with masked first and last elements
            in the selected rows."
  [map-vector
   rows
   number-of-columns]
  (if (empty? rows)
    map-vector
    (recur (mask-in-a-row map-vector
                          number-of-columns
                          (first rows)
                          [0 (dec number-of-columns)])
           (next rows)
           number-of-columns)))

(defn- mask-border
  "Masks cells, which belong to the map border.
      map-vector: A vector of map cell values.
      number-of-columns: A number of columns in the map.
   returns: Map with masked border values."
  [map-vector number-of-columns]
  (let [number-of-rows (/ (count map-vector) number-of-columns)]
    (-> map-vector
        (mask-row 0 number-of-columns)
        (mask-row (dec number-of-rows)
                  number-of-columns)
        (mask-first-and-last-elements (range number-of-rows)
                                      number-of-columns))))


;;; public methods

(defn mask-na-cells
  "Masks cells to nil values in case
        - a cell value is not defined,
        - or any value in the cell Moore neighbourhood is not defined,
        - or a cell belongs to the map-matrix border.
      map-vector: A vector of map cell values.
      number-of-columns: A number of columns in the map.
      na-values: List of na value positions per map row.
   returns: A list of cell values.

  ## Usage

      (require '[data-cleaning.filter :refer :all])

      (mask-na-cells [\"1\" \"2\" \"2\" \"3\" \"4\" \"1\"
                      \"3\" \"na\" \"3\" \"5\" \"6\" \"7\"
                      \"2\" \"8\" \"1\" \"2\" \"3\" \"1\"
                      \"1\" \"3\" \"2\" \"8\" \"1\" \"1\"
                      \"5\" \"4\" \"7\" \"na\" \"na\" \"1\"
                      \"6\" \"5\" \"1\" \"1\" \"1\" \"1\"]
                     6
                     {1 [1] 4 [3 4]})
      => [nil nil nil nil nil nil
          nil nil nil \"5\" \"6\" nil
          nil nil nil \"2\" \"3\" nil
          nil \"3\" nil nil nil nil
          nil \"4\" nil nil nil nil
          nil nil nil nil nil nil]

  "
  [map-vector
   number-of-columns
   na-values]
  (-> map-vector
      (mask-border number-of-columns)
      (mask-na number-of-columns na-values)))

(defn mask-only-na-cells
  "Masks na cells to nil values.
      map-vector: A vector of map cell values.
      number-of-columns: A number of columns in the map.
      na-values: List of na value positions per map row.
   returns: A list of cell values.

  ## Usage

      (require '[data-cleaning.filter :refer :all])

      (mask-only-na-cells [1 \"na\" 1 1 1
                           \"na\" \"na\" 1 1 1
                           1 1 1 \"na\" 1]
                          5
                          {0 [1] 1 [0 1] 2 [3]})
      => [1 nil 1 1 1
          nil nil 1 1 1
          1 1 1 nil 1]
  "
  [map-vector
   number-of-columns
   na-values]
  (if (empty? na-values)
    map-vector
    (recur (let [[row columns] (first na-values)]
             (mask-in-a-row map-vector
                            number-of-columns
                            row
                            columns))
           number-of-columns
           (rest na-values))))


