(ns data-cleaning.map
 (:require [clojure.string :as string]
           [utilities.reader :as reader]))


;;; private defs

;; Defines regular expression to delimit values from file lines.
(def ^:private delimiter-re #"\s")

;; Defines mask value.
(def ^:private mask 1)


;;; private methods

(defn- take-last-value
  "Returns a last value in the line.
      raw-data: File content.
      line-index: Line index.
   returns: Last value in the line."
  [raw-data line-index]
  (-> raw-data
      (nth line-index)
      (string/split delimiter-re)
      last))

(defn- take-value
  "Returns a last value (as a number) in the line.
      raw-data: File content.
      line-index: Line index.
      as-integer: A value indicating whether number should be parsed to integer type.
   returns: Last value in the line as number (double or integer)."
  ([raw-data line-index]
   (take-value raw-data
               line-index
               false))
  ([raw-data line-index as-integer]
   (let [value (take-last-value raw-data
                                line-index)]
     (if as-integer
       (Integer/parseInt value)
       (Double/parseDouble value)))))

(defn- try-to-parse-as-number
  "Tries to parse a last value in the line as number.
      raw-data: File content.
      line-index: Line index.
   returns: Last value in the line as number, if parsed successfully;
            otherwise, last value in the line as string."
  [raw-data line-index]
  (try (take-value raw-data
                   line-index)
    (catch NumberFormatException e
      (take-last-value raw-data
                       line-index))))

(defn- get-indexes
  "Returns positions, at which the specified value occurs in the collection.
      values: Collection of values.
      value: Value to search for.
   returns: Positions of the value in the collection."
  [values value]
  (reduce (fn [seed index]
            (let [current-value (nth values index)]
              (if (= (try (Double/parseDouble current-value)
                       (catch NumberFormatException e current-value))
                     value)
                (conj seed index)
                seed)))
          []
          (range (count values))))

(defn- map-matrix-to-properties
  "Returns maps properties with information about
        - number of elements in each row,
        - positions of NA values per row.
      map-matrix: Matrix with map values.
      na-value: Value, which is assumed to be 'non-applicable' cell values.
   returns: Map matrix properties."
  [map-matrix na-value]
  (reduce (fn [seed line]
            (let [values (string/split line delimiter-re)
                  updated-seed (update-in seed
                                          [:rows]
                                          #(conj % (count values)))
                  na-indexes (get-indexes values na-value)]
              (if (empty? na-indexes)
                (identity updated-seed)
                (update-in updated-seed
                           [:cells]
                           #(conj % {(count (get seed :rows)) na-indexes})))))
          (hash-map :rows []
                    :cells {})
          map-matrix))

(defn- parse-map-matrix
  "Recursively creates a list of map elements in the row-by-row order.
      lines: Lines of map elements.
      map-matrix: Accumulated list of map elements.
   returns: List of map elements in the row-by-row order."
  ([lines] (parse-map-matrix lines []))
  ([lines map-matrix]
  (if (empty? lines)
    map-matrix
    (recur (next lines)
           (into map-matrix
                 (string/split (first lines)
                               delimiter-re))))))


(defn- get-matrix
  "Returns a map matrix from .asc file content.
      raw-data: .asc file content.
   returns: A map matrix."
  [raw-data]
  (drop 6 raw-data))


;;; public methods

(defn read-properties
  "Returns .asc preamble data from the file with information about
        - map size (number of rows, number of columns) declared in .asc file,
        - number of elements in each row,
        - map resolution (cell size),
        - left bottom corner coordinates,
        - positions of NA values per row.
      map-path: Path to the map file.
   returns: .asc preamble data.

  ## Usage

      (require '[data-cleaning.map :refer :all])

      ;; \"raster.asc\"

      ;; ncols 5
      ;; nrows 3
      ;; xllcorner 1
      ;; yllcorner 2
      ;; cellsize 200
      ;; NODATA_value -9999
      ;; u 2 -9999 2
      ;; 4 6 -9999 4
      ;; 1
      ;; -9999 -9999 3 -9999 -9999
      ;; 1 2 3 4

      (read-properties \"raster.asc\")
      => {:declared-size [3 5]
          :rows [4 4 1 5 4]
          :cell-size) 200.0
          :left-bottom-corner [1.0 2.0]
          :na-cells {0 [2] 1 [2] 3 [0 1 3 4]}}
  "
  [map-path]
  (let [raw-data
        (reader/read-file map-path)
        content-properties
        (map-matrix-to-properties (get-matrix raw-data)
                                  (try-to-parse-as-number raw-data 5))]
    (hash-map :declared-size [(take-value raw-data 1 true)
                              (take-value raw-data 0 true)]
              :rows (:rows content-properties)
              :cell-size (take-value raw-data 4)
              :left-bottom-corner [(take-value raw-data 2)
                                   (take-value raw-data 3)]
              :na-cells (:cells content-properties))))

(defn read-map
  "Returns a map matrix from the .asc file.
      map-path: Path to the map file.
   returns: List of map elements in the row-by-row order.

  ## Usage

      (require '[data-cleaning.map :refer :all])

      ;; \"raster.asc\"

      ;; ncols 5
      ;; nrows 3
      ;; xllcorner 1
      ;; yllcorner 2
      ;; cellsize 200
      ;; NODATA_value -9999
      ;; u 2 -9999 2
      ;; 4 6 -9999 4
      ;; 1
      ;; -9999 -9999 3 -9999 -9999
      ;; 1 2 3 4

      (read-map \"raster.asc\")
      => [\"u\" \"2\" \"-9999\" \"2\"
          \"4\" \"6\" \"-9999\" \"4\"
          \"1\"
          \"-9999\" \"-9999\" \"3\" \"-9999\" \"-9999\"
          \"1\" \"2\" \"3\" \"4\"]
      "
  [map-path]
  (->> map-path
       reader/read-file
       get-matrix
       parse-map-matrix))

(defn read-positions
  "Returns positions for the masked cells from the .asc file.
      map-path: Path to the map file.
   returns: List of positions for masked cells.

  ## Usage

      (require '[data-cleaning.map :refer :all])

      ;; \"mask.asc\"

      ;; ncols 5
      ;; nrows 3
      ;; xllcorner 1
      ;; yllcorner 2
      ;; cellsize 200
      ;; NODATA_value -9999
      ;; u u -9999 u
      ;; 1 u -9999 1
      ;; u
      ;; -9999 -9999 u -9999 -9999
      ;; u u u 1.000

      (read-positions \"mask.asc\")
      => {1 [0 3] 4 [3]}
  "
  [map-path]
  (-> map-path
      reader/read-file
      get-matrix
      (map-matrix-to-properties (double mask))
      (:cells)))
