(ns data-cleaning.output-provider
   (:require [clojure.string :as string]
             [utilities.floating-point-comparison :refer :all]
             [utilities.format :as format]))


;;; private defs

;; Defines a value that is regarded as "not applicable" in the raster file.
(def ^:private na-value "-9999")

;; Defines a delimeter in the raster file.
(def ^:private delimeter " ")


;;; private methods

(defn- into-many-impl
  "Recursive implementation of method
   to successively conjoin collections.
      coll: Accumulated result.
      items: Collections object.
   returns: Successively conjoined collections."
  ([items] (into-many-impl [] items))
  ([coll items]
  (if (empty? items)
    coll
    (recur (into coll
                 (first items))
           (next items)))))

(defn- into-many
  "Returns successively conjoined collections.
      items: Collections object.
   returns: Successively conjoined collections."
  [& items]
  (into-many-impl items))

(defn- to-raster-preamble
  "Returns a collection of lines for the raster preamble data.
      map-properties: .asc preamble data.
   returns: Collection of lines for the raster preamble data."
  [map-properties]
  (let [get-property #(get map-properties %)]
    (into []
          (map #(string/join delimeter
                             [%1 %2])
               ["ncols"
                "nrows"
                "xllcorner"
                "yllcorner"
                "cellsize"
                "NODATA_value"]
               (into-many (vec (reverse (get-property :declared-size)))
                          (get-property :left-bottom-corner)
                          [(get-property :cell-size) na-value])))))

(defn- to-raster-line
  "Returns a line in the raster grid.
      items: Collection of values in the map row.
   returns: A line in the raster grid."
  [items]
  (string/join delimeter
               (map #(cond
                        (nil? %) na-value
                        (integer? %) (str %)
                        (number? %) (format/double-to-str (double %) 15)
                        :else %)
                    items)))

(defn- to-raster-grid
  "Returns a collection of lines for the raster grid.
      lines: Accumulated lines.
      map-vector: A vector of map cell values.
      number-of-columns: A number of columns in the map.
   returns: A collection of lines for the raster grid."
  [lines
   map-vector
   number-of-columns]
  (if (empty? map-vector)
    lines
    (recur (conj lines
                 (to-raster-line (take number-of-columns
                                       map-vector)))
           (drop number-of-columns
                 map-vector)
           number-of-columns)))

(defn- convert-to-message
  "Returns a list of strings with validation message and trace.
      message: Validation message.
      type: Validation message type.
   returns: A list of strings with validation message and trace."
  [message type]
  (cons (str type
             ": "
             (:message message))
        (map (fn [file-key]
               (str " file "
                    file-key
                    ":\n\t"
                    (get-in message
                            [:trace file-key])))
             ((comp sort
                    keys
                    :trace)
              message))))

(defn- to-cell-plot
  "Returns a tabular cell content for the specified neighbourhood.
      cells: A vector of values in the neighbourhood.
   returns: A tabular cell content with the neighbourhood."
  [cells]
  (->> cells
       (map #(str "\\textcolor{"
                  (if (= % 1) "black" "lightgray")
                  "}{\\rule{2mm}{2mm}}\\hspace{0.3mm}"))
       (partition 3)
       (map #(string/join (concat % " \\\\")))
       (cons " & \\begin{minipage}{7mm}\\baselineskip=0pt")
       (#(concat % ["\\end{minipage} "]))))

(defn- traverse
  "Traverses through the hash-map with Moore neighbourhood statistics.
      statistics: A hash-map with Moore neighbourhood statistics.
      convert: Method to convert a statistical node.
   returns: Converted Moore neighbourhood statistics."
  [statistics
   convert]
  (->> statistics
       (sort-by second >)
       ((fn [combinations]
          (map (fn [[combination frequency]
                    index
                    [down up]]
                 (convert combination
                          frequency
                          index
                          (if (= (+ down up) 0)
                            index
                            (str (- index down)
                                 "-"
                                 (+ index up)))))
               combinations
               (->> (count combinations)
                    inc
                    (range 1))
               (->> combinations
                    (map second)
                    (partition-by identity)
                    (map #(map (fn[index]
                                 [(dec index)
                                  (- (count %)
                                     index)])
                               (range 1 (inc (count %)))))
                    (reduce #(concat %1 %2)
                            '())))))))

(defn- to-percent
  "Returns a value in the percent format.
      total-frequency: A value of total frequency.
      value: Value object.
   returns: A value in the percent format."
  [total-frequency value]
  (str (->> value
            (#(if (= 0 total-frequency)
                0
                (/ % total-frequency)))
            (* 100)
            double
            (#(format/double-to-str % 2)))
       "%"))


;;; public methods

(defn map-to-raster-file
  "Returns a collection of the raster file lines.
      map-vector: A vector of map cell values.
      map-properties: .asc preamble data.
   returns: Collection of the raster file lines.

  ## Usage

      (require '[data-cleaning.output-provider :refer :all])

      (map-to-raster-file [1 nil 2 nil 1 1.1 1 7 nil 1 \"1.0\" 1.0]
                          {:declared-size [3 4]
                           :cell-size 200.0
                           :left-bottom-corner [1.0 2.0]})
      => [\"ncols 4\"
          \"nrows 3\"
          \"xllcorner 1.0\"
          \"yllcorner 2.0\"
          \"cellsize 200.0\"
          \"NODATA_value -9999\"
          \"1 -9999 2 -9999\"
          \"1 1.100000000000000 1 7\"
          \"-9999 1 1.0 1.000000000000000\"]
  "
  [map-vector
   map-properties]
  (-> (to-raster-preamble map-properties)
      (to-raster-grid map-vector
                      (second (:declared-size map-properties)))))

(defn convert-to-error-message
  "Returns a list of strings with validation error message and trace.
      error: Validation error.
   returns: A list of strings with error message and trace.

  ## Usage

      (require '[data-cleaning.output-provider :refer :all])

      (convert-to-error-message {:message \"error message content\"
                                 :trace {\"name1\"
                                         {:rows 1 :columns [1 2 3]}
                                         \"name2\"
                                         2}})
      => '(\"Error: error message content\"
           \" file name1:\\n\\t{:rows 1, :columns [1 2 3]}\"
           \" file name2:\\n\\t2\")
  "
  [error]
  (convert-to-message error "Error"))

(defn convert-to-warning-message
  "Returns a list of strings with validation warning message and trace.
      warning: Validation warning.
   returns: A list of strings with warning message and trace.

  ## Usage

      (require '[data-cleaning.output-provider :refer :all])

      (convert-to-warning-message {:message \"warning message content\"
                                   :trace {\"name1\"
                                           {:rows 1 :columns [1 2 3]}
                                           \"name2\"
                                           2}})
      => '(\"Warning: warning message content\"
           \" file name1:\\n\\t{:rows 1, :columns [1 2 3]}\"
           \" file name2:\\n\\t2\")
  "
  [warning]
  (convert-to-message warning "Warning"))

(defn stat-to-csv
  "Returns a csv content with frequency
   for each Moore neighbourhood.
      statistics: A hash-map with Moore neighbourhood statistics.
   returns: A csv content.

  ## Usage

      (require '[data-cleaning.output-provider :refer :all])

      (stat-to-csv {[9 8 7 6 5 4 3 2 1] 2
                    [1 2 3 4 5 6 7 8 9] 3
                    [1 1 1 1 1 1 1 1 1] 1
                    [0 0 0 0 0 0 0 0 0] 2})
      => '([\"No.\"
            \"Left-upper\" \"Center-upper\" \"Right-upper\"
            \"Left-middle\" \"Center-middle\" \"Right-middle\"
            \"Left-lower\" \"Center-lower\" \"Right-lower\"
            \"Rank\" \"Frequency\"]
           [\"1\" \"1\" \"2\" \"3\" \"4\" \"5\" \"6\" \"7\" \"8\" \"9\" \"1\" \"3\"]
           [\"2\" \"9\" \"8\" \"7\" \"6\" \"5\" \"4\" \"3\" \"2\" \"1\" \"2-3\" \"2\"]
           [\"3\" \"0\" \"0\" \"0\" \"0\" \"0\" \"0\" \"0\" \"0\" \"0\" \"2-3\" \"2\"]
           [\"4\" \"1\" \"1\" \"1\" \"1\" \"1\" \"1\" \"1\" \"1\" \"1\" \"4\" \"1\"])
  "
  [statistics]
  (->> (traverse statistics
                 (fn [combination
                      frequency
                      index
                      rank]
                   (->> (concat [index]
                                combination
                                [rank]
                                [frequency])
                        (map str)
                        vec)))
       (cons ["No."
              "Left-upper" "Center-upper" "Right-upper"
              "Left-middle" "Center-middle" "Right-middle"
              "Left-lower" "Center-lower" "Right-lower"
              "Rank" "Frequency"])))

(defn stat-to-latex-tabular
  "Returns a latex tabular content with frequency
   for each Moore neighbourhood.
      statistics: A hash-map with Moore neighbourhood statistics.
   returns: A content of latex tabular section.

  ## Usage

      (require '[data-cleaning.output-provider :refer :all])

      (stat-to-latex-tabular {[1 1 0 0 1 0 0 0 1] 3})
      => [\"1\"
          \" & \\begin{minipage}{7mm}\\baselineskip=0pt\"
          \"\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\\"
          \"\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\\"
          \"\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\\"
          \"\\end{minipage} \"]
  "
  [statistics]
  (->> (traverse statistics
                 (fn [combination
                      frequency
                      index
                      rank]
                   (->> (vector rank frequency)
                        (map #(str "& "
                                   %
                                   " "))
                        reverse
                        (into (list "\\vspace{2mm} \\\\"))
                        string/join
                        vector
                        (concat [(str index)]
                                (to-cell-plot combination)))))
       flatten
       vec))
