(ns data-cleaning.spatial-maps)


;;; private defs

;; Defines an urban cell value.
(def ^:private urban-cell "1")

;;; private methods

(defn- classify-as-binary
  "Perfoms binary classification.
      cell-value: Cell value.
   returns: 1, if cell value is of the urban type;
            0, otherwiise."
  [cell-value]
  (if (= cell-value urban-cell) 1 0))

(defn- create-layer
  "Creates a binary layer with values
   of the specified cell neighbour.
      map-vector: A vector of map cell values.
      get-relative-cell-value: Method to get a value of the neighbour cell.
   returns: A binary layer with values of the cell neighbour."
  [map-vector
   get-relative-cell-value]
  (into []
        (map (fn [cell-value position]
               (if (nil? cell-value)
                 nil
                 (classify-as-binary (get-relative-cell-value position))))
             map-vector
             (range (count map-vector)))))

(defn- to-double
  "Converts map values to the double format.
      map-vector: A vector of map cell values.
   returns: A list of double cell values."
  [map-vector]
  (map #(if (not (nil? %))
          (Double/parseDouble %))
       map-vector))


;;; public methods

(defn form-binary-layers-for-neighbour-cells
  "Returns binary maps by the specified cell type
   for type of cells in the Moore neighbourhood (8 maps).
      map-vector: A vector of map cell values (filtered).
      origin: A vector of original map cell values.
      number-of-columns: A number of columns in the map.
   returns: Collection of binary maps for types of each neighbour cell.

  ## Usage

      (require '[data-cleaning.spatial-maps :refer :all])

      (land-use-map [nil nil nil nil nil nil \"1\" 2 nil nil nil \"1\" \"1\" nil nil
                     nil \"1\" \"1\" nil nil nil 2 2 3 nil nil \"1\" nil \"1\" nil
                     nil nil nil nil nil])
      (origin [2 2 \"1\" \"1\" 2 \"1\" \"1\" 2 2 nil 2 \"1\" \"1\" \"1\" nil
               3 \"1\" \"1\" \"1\" \"1\" 4 2 2 3 \"1\" 5 \"1\" 2 \"1\" \"1\"
               \"1\" \"1\" \"1\" 2 3])

      (form-binary-layers-for-neighbour-cells land-use-map
                                              origin
                                              5)
      => [; left-upper-neighbour
          [nil nil nil nil nil nil 0 0 nil nil nil 1 1 nil nil nil 0 1 nil nil
           nil 0 1 1 nil nil 0 nil 0 nil nil nil nil nil nil]
          ; center-upper-neighbour
          [nil nil nil nil nil nil 0 1 nil nil nil 1 0 nil nil nil 1 1 nil nil
           nil 1 1 1 nil nil 0 nil 0 nil nil nil nil nil nil]
          ; right-upper-neighbour
          [nil nil nil nil nil nil 1 1 nil nil nil 0 0 nil nil nil 1 1 nil nil
           nil 1 1 1 nil nil 0 nil 1 nil nil nil nil nil nil]
          ; left-middle-neighbour
          [nil nil nil nil nil nil 1 1 nil nil nil 0 1 nil nil nil 0 1 nil nil
           nil 0 0 0 nil nil 0 nil 0 nil nil nil nil nil nil]
          ; center-middle-neighbour
          [nil nil nil nil nil nil 1 0 nil nil nil 1 1 nil nil nil 1 1 nil nil
           nil 0 0 0 nil nil 1 nil 1 nil nil nil nil nil nil]
          ; right-middle-neighbour
          [nil nil nil nil nil nil 0 0 nil nil nil 1 1 nil nil nil 1 1 nil nil
           nil 0 0 1 nil nil 0 nil 1 nil nil nil nil nil nil]
          ; left-lower-neighbour
          [nil nil nil nil nil nil 0 1 nil nil nil 0 1 nil nil nil 0 0 nil nil
           nil 0 1 0 nil nil 1 nil 1 nil nil nil nil nil nil]
          ; center-lower-neighbour
          [nil nil nil nil nil nil 1 1 nil nil nil 1 1 nil nil nil 0 0 nil nil
           nil 1 0 1 nil nil 1 nil 0 nil nil nil nil nil nil]
          ; right-lower-neighbour
          [nil nil nil nil nil nil 1 1 nil nil nil 1 1 nil nil nil 0 0 nil nil
           nil 0 1 1 nil nil 1 nil 0 nil nil nil nil nil nil]]
  "
  [map-vector
   origin
   number-of-columns]
  (map (fn [get-position]
         (create-layer map-vector
                       #(nth origin (get-position %))))
       [; get left upper cell position
        #(dec (- % number-of-columns))
        ; get center upper cell position
        #(- % number-of-columns)
        ; get right upper cell position
        #(inc (- % number-of-columns))
        ; get left cell position
        #(dec %)
        ; get center cell position
        #(identity %)
        ; get right cell position
        #(inc %)
        ; get left lower cell position
        #(dec (+ % number-of-columns))
        ; get center lower cell position
        #(+ % number-of-columns)
        ; get right lower cell position
        #(inc (+ % number-of-columns))]))

(defn transform
  "Transforms cell values with the specified operation.

   Supported operations:
      - :log (x'=ln(x)),
      - :unit-rescaling (x'=(x-x_min)/(x_max-x_min).

   If operation-id is not known, original values are returned
   (from the filtered map).

      cleaned-map: A list of cell values in the already filtered map.
      original-map: A list of original cell values.
      operation-id: An identifier of transformation.

   returns: A list of transformed values.

   ## Usage

      (require '[data-cleaning.spatial-maps :refer :all])

      (transform [nil nil nil nil \"1\" \"2\"]
                 [nil nil nil \"3\" \"1\" \"2\"]
                 :unit-rescaling)
      => [nil nil nil nil 0.0 0.5]

      (transform [nil nil nil nil \"1\" \"2\"]
                 [nil nil nil nil \"1\" \"2\"]
                 :log)
      => [nil nil nil nil 0.0 0.6931471805599453]

      (transform [nil nil nil nil \"1\" \"2\"]
                 [nil nil nil nil \"1\" \"2\"]
                 :unknown)
      => [nil nil nil nil \"1\" \"2\"]

   "
  [cleaned-map
   original-map
   operation-id]
  (let [double-map (to-double cleaned-map)]
    (->> (condp = operation-id
           :log (partial map
                         #(if (not (nil? %))
                            (Math/log %)))
           :unit-rescaling (fn[values]
                             (->> (to-double original-map)
                                  (filter #(not (nil? %)))
                                  ((juxt #(apply min %)
                                         #(apply max %)))
                                  ((fn[[x_min x_max]]
                                     ;(do
                                      ; (println x_min)
                                       ;(println x_max)
                                     (map #(if (not (nil? %))
                                             (-> (- % x_min)
                                                 (/ (- x_max x_min))))
                                          values)))))
           (fn[_](identity cleaned-map)))
         (#(% double-map))
         (into []))))
