(ns data-cleaning.output-provider-test
  (:require [clojure.test :refer :all]
            [data-cleaning.output-provider :refer :all]))

(deftest map-to-raster-file-test
  (testing "Map data should be converted to the raster file format."
    (
     ; Arrange
     let [map-vector1 ["1" nil "2" nil
                       "2" "4" "4" "7"
                       nil "1" "1" "1.0"]
          map-properties (hash-map :declared-size [3 4]
                                   :cell-size 200.0
                                   :left-bottom-corner [1.0 2.0])
          map-vector2 [1 nil 2 nil
                       1 1.1 1 7
                       nil 1 "1.0" 1.0]]

       ; Act
       (let [raster-file1 (map-to-raster-file map-vector1
                                              map-properties)
             raster-file2 (map-to-raster-file map-vector2
                                              map-properties)]

         ; Assert
         (is (= raster-file1
                ["ncols 4"
                 "nrows 3"
                 "xllcorner 1.0"
                 "yllcorner 2.0"
                 "cellsize 200.0"
                 "NODATA_value -9999"
                 "1 -9999 2 -9999"
                 "2 4 4 7"
                 "-9999 1 1 1.0"]))

         (is (= raster-file2
                ["ncols 4"
                 "nrows 3"
                 "xllcorner 1.0"
                 "yllcorner 2.0"
                 "cellsize 200.0"
                 "NODATA_value -9999"
                 "1 -9999 2 -9999"
                 "1 1.100000000000000 1 7"
                 "-9999 1 1.0 1.000000000000000"]))))))

(deftest convert-to-error-message-test
  (testing "Validation error should be converted to a sequence of lines."
    (
     ; Act
     let [lines (convert-to-error-message {:message "error message content"
                                           :trace {"name1" {:rows 1 :columns [1 2 3]}
                                                   "name2" 2}})]

       ; Assert
       (is (= lines
              '("Error: error message content"
               " file name1:\n\t{:rows 1, :columns [1 2 3]}"
               " file name2:\n\t2"))))))

(deftest convert-to-warning-message-test
  (testing "Validation warning should be converted to a sequence of lines."
    (
     ; Act
     let [lines (convert-to-warning-message {:message "warning message content"
                                             :trace {"name1" {:rows 1 :columns [1 2 3]}
                                                     "name2" 2}})]

       ; Assert
       (is (= lines
              '("Warning: warning message content"
               " file name1:\n\t{:rows 1, :columns [1 2 3]}"
               " file name2:\n\t2"))))))

(deftest stat-to-csv-test
  (testing "Summary statistics by each land-use combination should be converted to csv file format."
    (
     ; Act
     let [csv (stat-to-csv {[9 8 7 6 5 4 3 2 1] 2
                            [1 2 3 4 5 6 7 8 9] 3
                            [1 1 1 1 1 1 1 1 1] 1
                            [0 0 0 0 0 0 0 0 0] 2})]
       ; Assert
       (is (= csv
              '(["No."
                 "Left-upper" "Center-upper" "Right-upper"
                 "Left-middle" "Center-middle" "Right-middle"
                 "Left-lower" "Center-lower" "Right-lower"
                 "Rank" "Frequency"]
                ["1" "1" "2" "3" "4" "5" "6" "7" "8" "9" "1" "3"]
                ["2" "9" "8" "7" "6" "5" "4" "3" "2" "1" "2-3" "2"]
                ["3" "0" "0" "0" "0" "0" "0" "0" "0" "0" "2-3" "2"]
                ["4" "1" "1" "1" "1" "1" "1" "1" "1" "1" "4" "1"]))))))

(deftest stat-to-latex-tabular-test
  (testing "Frequency statistics by each land-use combination should be converted to rows in latex tabular section."
    (
     ; Act
     let [rows (stat-to-latex-tabular {[1 1 1 1 1 1 1 1 1] 2
                                       [1 1 0 0 1 0 0 0 1] 3
                                       [0 0 0 0 0 0 0 0 0] 1})]
       ; Assert
       (is (= rows
              ["1"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 1 & 3 \\vspace{2mm} \\\\"
               "2"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{black}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 2 & 2 \\vspace{2mm} \\\\"
               "3"
               " & \\begin{minipage}{7mm}\\baselineskip=0pt"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm}\\textcolor{lightgray}{\\rule{2mm}{2mm}}\\hspace{0.3mm} \\\\"
               "\\end{minipage} "
               "& 3 & 1 \\vspace{2mm} \\\\"])))))
