(ns data-cleaning.sampling-test
  (:require [clojure.test :refer :all]
            [data-cleaning.sampling :refer :all]
            [utilities.floating-point-comparison :refer :all]))


;;; tests

(deftest clean-and-count-urban-cells
  (testing "Land use map should contain number of urban cells in the Moore neighbourhood."
    (
     ; Act
     let [sample (clean-and-transform {:na-cells {5 [3], 2 [0 5 6], 1 [0 5 6], 0 [4 5 6]},
                                       :declared-size [6 7]}
                                      {:land-use
                                       "test/data_cleaning/test_files/test_case/land_use.asc"})]

       ; Assert
       (is (= (:land-use-matrix sample)
              ["land_use.asc"
               [nil nil nil nil nil nil nil
                nil nil 4 nil nil nil nil
                nil nil 5 4 nil nil nil
                nil nil 5 4 nil nil nil
                nil 2 nil nil nil 2 nil
                nil nil nil nil nil nil nil]])))))

(deftest clean-and-return-neighbours
  (testing "Matrices of the cells in the Moore neighbourhood should be created."
    (
     ; Act
     let [sample (clean-and-transform {:na-cells {5 [3], 2 [0 5 6], 1 [0 5 6], 0 [4 5 6]},
                                       :declared-size [6 7]}
                                      {:land-use
                                       "test/data_cleaning/test_files/test_case/land_use.asc"})]

       ; Assert
       (is (= (:neighbour-matrices sample)
              [[nil nil nil nil nil nil nil
                nil nil 0 nil nil nil nil
                nil nil 0 0 nil nil nil
                nil nil 0 1 nil nil nil
                nil 0 nil nil nil 0 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 1 nil nil nil nil
                nil nil 0 1 nil nil nil
                nil nil 1 1 nil nil nil
                nil 1 nil nil nil 0 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 0 nil nil nil nil
                nil nil 1 0 nil nil nil
                nil nil 1 0 nil nil nil
                nil 0 nil nil nil 0 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 0 nil nil nil nil
                nil nil 0 1 nil nil nil
                nil nil 1 0 nil nil nil
                nil 0 nil nil nil 1 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 0 nil nil nil nil
                nil nil 1 1 nil nil nil
                nil nil 0 1 nil nil nil
                nil 1 nil nil nil 0 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 1 nil nil nil nil
                nil nil 1 0 nil nil nil
                nil nil 1 0 nil nil nil
                nil 0 nil nil nil 0 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 0 nil nil nil nil
                nil nil 1 0 nil nil nil
                nil nil 1 0 nil nil nil
                nil 0 nil nil nil 0 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 1 nil nil nil nil
                nil nil 0 1 nil nil nil
                nil nil 0 0 nil nil nil
                nil 0 nil nil nil 0 nil
                nil nil nil nil nil nil nil]
               [nil nil nil nil nil nil nil
                nil nil 1 nil nil nil nil
                nil nil 1 0 nil nil nil
                nil nil 0 1 nil nil nil
                nil 0 nil nil nil 1 nil
                nil nil nil nil nil nil nil]])))))

(deftest clean-and-transform-maps
  (testing "Maps should be filtered and transformed if necessary."
    (
     ; Act
     let [sample (clean-and-transform {:na-cells {0 [0]}
                                       :declared-size [5 4]}
                                      {:land-use
                                       "test/data_cleaning/test_files/land_use.asc"
                                       :other
                                       {"test/data_cleaning/test_files/var_self.asc"
                                        :identity
                                        "test/data_cleaning/test_files/var_unit.asc"
                                        :unit-rescaling
                                        "test/data_cleaning/test_files/var_log.asc"
                                        :log}})
          unit-map (get-in sample [:var-matrices "var_unit.asc"])]

       ; Assert
       (is (= (get-in sample [:var-matrices "var_self.asc"])
              [nil nil nil nil
               nil nil "2" nil
               nil "1" "1" nil
               nil "1" "1.4" nil
               nil nil nil nil]))

       (is (= (get-in sample [:var-matrices "var_log.asc"])
              [nil nil nil nil
               nil nil (Math/log 2) nil
               nil 0.0 0.0 nil
               nil 0.0 (Math/log 1.4) nil
               nil nil nil nil]))

       (is (= (count unit-map)
              20))
       (is (= (nth unit-map 0)
              (nth unit-map 1)
              (nth unit-map 2)
              (nth unit-map 3)
              (nth unit-map 4)
              (nth unit-map 5)
              nil))
       (is (double= (nth unit-map 6)
                    0.5))
       (is (= (nth unit-map 7)
              nil))
       (is (= (nth unit-map 8)
              nil))
       (is (double= (nth unit-map 9)
                    0.0))
       (is (double= (nth unit-map 10)
                    0.0))
       (is (= (nth unit-map 11)
              nil))
       (is (= (nth unit-map 12)
              nil))
       (is (double= (nth unit-map 13)
                    0.0))
       (is (double= (nth unit-map 14)
                    0.9))
       (is (= (nth unit-map 15)
              (nth unit-map 16)
              (nth unit-map 17)
              (nth unit-map 18)
              (nth unit-map 19)
              nil)))))

(deftest transform-with-original-map-where-only-own-na-values-excluded
  (testing "The map without only own na values should be used in transformation"
    (
     ; Act
     let [unit-map (-> (clean-and-transform {:na-cells {0 [0] 4 [0]}
                                             :declared-size [5 4]}
                                             {:land-use
                                              "test/data_cleaning/test_files/land_use.asc"
                                              :other
                                              {"test/data_cleaning/test_files/var_unit.asc"
                                               :unit-rescaling}})
                       (get-in [:var-matrices "var_unit.asc"]))]

      ; Assert
      (is (= (count unit-map)
             20))
         (is (= (nth unit-map 0)
                (nth unit-map 1)
                (nth unit-map 2)
                (nth unit-map 3)
                (nth unit-map 4)
                (nth unit-map 5)
                nil))
         (is (double= (nth unit-map 6)
                      0.5))
         (is (= (nth unit-map 7)
                (nth unit-map 8)
                nil))
         (is (double= (nth unit-map 9)
                      0.0))
         (is (double= (nth unit-map 10)
                      0.0))
         (is (= (nth unit-map 11)
                (nth unit-map 12)
                (nth unit-map 13)
                nil))
         (is (double= (nth unit-map 14)
                      0.9))
         (is (= (nth unit-map 15)
                (nth unit-map 16)
                (nth unit-map 17)
                (nth unit-map 18)
                (nth unit-map 19)
                nil)))))


;;; test grouping

(deftest clean-and-transform-test
  (testing "Map matrices should be cleaned and transformed if necessary:\n"
    (clean-and-count-urban-cells)
    (clean-and-return-neighbours)
    (clean-and-transform-maps)
    (transform-with-original-map-where-only-own-na-values-excluded)))

;;; tests in the namespace

(defn test-ns-hook
  "summary: Explicit definition of tests in the namespace.
   returns: Explicit definition of tests in the namespace."
  []
  (clean-and-transform-test))
