(ns data-cleaning.validation-test
  (:require [clojure.test :refer :all]
            [data-cleaning.validation :refer :all]))

;;; tests

(deftest nil-should-be-returned-if-validation-was-passed-successfully
  (testing "If validation was passed successfully,
    nil error value and empty warning list should be returned."
    (
     ;Act
     let [result (validate-maps {"1.asc" {:declared-size [1 1]
                                          :rows [1]
                                          :cell-size 1.0
                                          :left-bottom-corner [1.0 1.0]
                                          :na-cells {}}})]

     ; Assert
     (is (= (get result :error)
            nil))
     (is (= (get result :warnings)
            '())))))

(deftest validate-declared-map-sizes
  (testing "Validation error should be returned,
    if a raster map size doesn't fit with data."
    (
     ; Arrange
     let [maps {"1.asc" {:declared-size [1 1]
                         :rows [1 1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "2.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "3.asc" {:declared-size [3 2]
                         :rows [2 1 2]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}}]

       ; Act
       (let [error (get (validate-maps maps) :error)]

         ; Assert
         (is (= (get error :message)
                "Declared number of rows and columns doesn't fit data."))
         (is (= (get error :trace)
                {"1.asc" {:rows 2 :columns [1 1]}
                 "3.asc" {:rows 3 :columns [2 1 2]}}))))))

(deftest validate-equality-of-map-sizes
  (testing "Validation error should be returned,
    if maps have different number of rows or columns."
    (
     ; Arrange
     let [maps {"1.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "2.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "3.asc" {:declared-size [1 2]
                         :rows [2]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}}]

       ; Act
       (let [error (get (validate-maps maps) :error)]

         ; Assert
         (is (= (get error :message)
                "Grid declaration is different in the source maps."))
         (is (= (get error :trace)
                {"1.asc" {:rows 1 :columns 1 :cell-size 1.0 :left-bottom-corner [1.0 1.0]}
                 "2.asc" {:rows 1 :columns 1 :cell-size 1.0 :left-bottom-corner [1.0 1.0]}
                 "3.asc" {:rows 1 :columns 2 :cell-size 1.0 :left-bottom-corner [1.0 1.0]}}))))))

(deftest validate-equality-of-map-resolutions
  (testing "Validation error should be returned,
    if maps have different resolution."
    (
     ; Arrange
     let [maps {"1.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "2.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 2.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "3.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}}]

       ; Act
       (let [error (get (validate-maps maps) :error)]

         ; Assert
         (is (= (get error :message)
                "Grid declaration is different in the source maps."))
         (is (= (get error :trace)
                {"1.asc" {:rows 1 :columns 1 :cell-size 1.0 :left-bottom-corner [1.0 1.0]}
                 "2.asc" {:rows 1 :columns 1 :cell-size 2.0 :left-bottom-corner [1.0 1.0]}
                 "3.asc" {:rows 1 :columns 1 :cell-size 1.0 :left-bottom-corner [1.0 1.0]}}))))))

(deftest validate-equality-of-map-coordinates
  (testing "Validation error should be returned,
    if maps have different referenced coordinates."
    (
     ; Arrange
     let [maps {"1.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "2.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {}}
                "3.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 2.0]
                         :na-cells {}}}]

       ; Act
       (let [error (get (validate-maps maps) :error)]

         ; Assert
         (is (= (get error :message)
                "Grid declaration is different in the source maps."))
         (is (= (get error :trace)
                {"1.asc" {:rows 1 :columns 1 :cell-size 1.0 :left-bottom-corner [1.0 1.0]}
                 "2.asc" {:rows 1 :columns 1 :cell-size 1.0 :left-bottom-corner [1.0 1.0]}
                 "3.asc" {:rows 1 :columns 1 :cell-size 1.0 :left-bottom-corner [1.0 2.0]}}))))))

(deftest validate-equality-of-na-value-positions
  (testing "Validation warning should be returned,
    if maps have different positions of NA values."
    (
     ; Arrange
     let [maps {"1.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {0 [1] 1 [1 2] 2 [1]}}
                "2.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {0 [1] 1 [2] 2 [1]}}
                "3.asc" {:declared-size [1 1]
                         :rows [1]
                         :cell-size 1.0
                         :left-bottom-corner [1.0 1.0]
                         :na-cells {0 [1] 1 [1 2] 2 [1]}}}]

       ; Act
       (let [warnings (get (validate-maps maps) :warnings)]

         ; Assert
         (is (= (count warnings)
                1))
         (is (= (get (first warnings) :message)
                "NA value positions are different in the source maps."))
         (is (= (get (first warnings) :trace)
                {"1.asc" 4 "2.asc" 3 "3.asc" 4}))))))


;;; test grouping

(deftest validate-maps-test
  (testing "Maps should be validated for consistency:\n"
    (nil-should-be-returned-if-validation-was-passed-successfully)
    (validate-declared-map-sizes)
    (validate-equality-of-map-sizes)
    (validate-equality-of-map-resolutions)
    (validate-equality-of-map-coordinates)
    (validate-equality-of-na-value-positions)))


;;; tests in the namespace

(defn test-ns-hook
  "summary: Explicit definition of tests in the namespace.
   returns: Explicit definition of tests in the namespace."
  []
  (validate-maps-test))
