# Regression-tests

Regression-tests runs multiple linear regression with the unknown joint distribution of error terms
using OLS and resampling.

This includes: 1) hypothesis testing using permutations method
                  (model and coefficents significance),
               2) estimating confidence intervals for regression model parameters
                  using bootstrapping,
               3) estimating approximation accuracy using bootstrapping,
               4) bootstrap hypothesis testing on spatial autocorrelation.


## Libraries

* csv - permutation testing and bootstrapping with a csv formatted output
* permutations - runs permutation tests
* bootstrap - calculates bootstrap estimates and runs bootstrap
              hypothesis testing for spatial autocorrelation property
* sample-tests - calculates spatial autocorrelation indexes

## Install

No installation needed. This is a standalone java application.
Requires JRE 1.8.

To include libraries add the following to your `:dependencies`:

### Leiningen
```clj
[org.iiasa/regression-tests "1.0.0"]
```

### Maven

```xml
<dependency>
  <groupId>org.iiasa</groupId>
  <artifactId>regression-tests</artifactId>
  <version>1.0.0</version>
</dependency>
```

## Options

    -t, --trace   Print stack trace
    -h, --help    Print command help

## Usage

examples:

`````
(require '[regression-tests.csv :refer :all])

(run-permutation-tests sample-x1-x2-x3.csv
                       (fn[coll] (vec (repeatedly 10000 #(shuffle coll)))))

(require 'incanter.stats) ;;; incanter version 1.5.7
(bootstrap-regression sample-x1-x2-x3.csv
                      (fn[coll](vec (repeatedly 10000  #(incanter.stats/sample coll)))))
(bootstrap-accuracy sample-x1-x2-x3.csv
                    cells-x1-x2-x3.csv
                    (fn[coll](vec (repeatedly 10000  #(incanter.stats/sample coll)))))
(bootstrap-independence-tests sample-x1-x2-x3.csv
                              neighbours-matrix.csv
                              (fn[coll](vec (repeatedly 10000  #(incanter.stats/sample coll)))))
`````

`````
(def y [1.1 1 0.95 1.15 2.1 2.05 3 3.01 3.02 2.9])
(def x [[1 1 4] [2 3 2] [2 2 3] [1.5 1.5 1.5] [3 3.1 5] [3.5 3 5.5]
        [4 3 6] [3.8 2.5 6.3] [3.9 2.7 6.5] [4.2 3.4 6]])

(def permutations [[0 2 1 8 3 4 5 6 9 7]
                   [5 0 1 3 2 4 8 6 7 9]
                   [0 2 3 5 4 1 6 7 8 9]])
(def with-replacement [[0 1 2 0 4 5 6 7 8 6]
                       [0 1 5 3 5 5 6 7 0 9]
                       [0 1 2 3 5 5 6 7 8 9]
                       [0 1 2 3 4 5 8 7 8 9]])
```````
(require '[regression-tests.permutations :refer :all])

`````````
Use 'permute-r-squared' to test null hypothesis that
there is no linear relationship between y and x (overall model significance).

(permute-r-squared y x permutations)
=>
    {:p-value [0.25 -0.17435244785437493 0.674352447854375],
     :sample [0.9467146097039536 0.7483687115380364 0.7974802262317389 0.6996691908442358]}

`````````
Use 'permute-t-stat' to test significance of individual coefficients [Freedman & Lane, 1983].

(permute-t-stat y x permutations 0)
=> [0.0 0.0 0.0]
(permute-t-stat y x permutations 1)
=> [0.0 0.0 0.0]
(permute-t-stat y x permutations 2)
=> [0.25 -0.17435244785437493 0.674352447854375]

```````
(require '[regression-tests.bootstrap :refer :all])

`````````
Use 'regression-stats-percentiles' to estimate confidence intervals
with the percentile bootstrap scheme (Efron & Tibshirani, 1993).

(regression-stats-percentiles y x with-replacement)
=>
   {:r-squared [0.9401664271263236 0.9573638262472198 0.9485972809614914],
    :mse [0.05119649671254231 0.07484392157787506 0.0625135321898773],
    :coefficients [[0.5095454591100292 1.0084063307610052 0.8108528980792314]
                   [-0.6364733499450175 -0.11761529878943122 -0.41016539093890503]
                   [-0.014289379521162715 0.26277051757180914 0.09815780110702263]
                   [-0.3872946185037165 0.7366170749180654 0.23879206650943702]]}

`````````
Use 'accuracy-percentiles' to estimate confidence intervals of approximation quantiles.

(def cells {:y [1 1.4 3.4]
            :urban [1 2 1]
            :x [[1 1 1] [1.1 1.3 1.1] [3 4 1]]})

(accuracy-percentiles y x cells with-replacement)
=>
    {:sample {:p-min [0.7325939406113093 0.09426067621289036 0.1313719684139727 0.2726825935013899 0.2694252999022879],
              :p-25-percent [0.7325939406113093 0.09426067621289036 0.1313719684139727 0.2726825935013899 0.2694252999022879],
              :p-50-percent [1.0906469325799546 0.39726963364663037 0.5991781261463421 0.694437572706074 0.6910235979522977],
              :p-75-percent [2.4663489187595444 2.1983467121001516 2.271866486743393 2.2249032313817714 2.194299660522469],
              :p-max [2.4663489187595444 2.1983467121001516 2.271866486743393 2.2249032313817714 2.194299660522469]},
     :p-value {["all" :p-min] [0.09426067621289036 0.7325939406113093 0.30006689572837003],
               ["all" :p-25-percent] [0.09426067621289036 0.7325939406113093 0.30006689572837003],
               ["all" :p-50-percent] [0.39726963364663037 1.0906469325799546 0.6945111726062597],
               ["all" :p-75-percent] [2.194299660522469 2.4663489187595444 2.271153001901466],
               ["all" :p-max] [2.194299660522469 2.4663489187595444 2.271153001901466],

               [1 :p-min] [0.09426067621289036 0.7325939406113093 0.30006689572837003],
               [1 :p-25-percent] [0.09426067621289036 0.7325939406113093 0.30006689572837003],
               [1 :p-75-percent] [2.194299660522469 2.4663489187595444 2.271153001901466],
               [1 :p-50-percent] [0.09426067621289036 0.7325939406113093 0.30006689572837003],
               [1 :p-max] [2.194299660522469 2.4663489187595444 2.271153001901466],

               [2 :p-min] [0.39726963364663037 1.0906469325799546 0.6945111726062597],
               [2 :p-25-percent] [0.39726963364663037 1.0906469325799546 0.6945111726062597],
               [2 :p-50-percent] [0.39726963364663037 1.0906469325799546 0.6945111726062597],
               [2 :p-75-percent] [0.39726963364663037 1.0906469325799546 0.6945111726062597],
               [2 :p-max] [0.39726963364663037 1.0906469325799546 0.6945111726062597]}}

`````````
Use 'residuals-independence' to perform bootstrap hypothesis testing [Efron & Tibshirani, 1993; Lin et al., 2011]
to check spatial autocorrelation property.

Global spatial autocorrelation indexes - Moran's I [Moran, 1950] and Geary's C [Geary, 1954].

(def neighbours-matrix {0 [1 5] 1 [0] 3 [9] 5 [0 7 8 9] 7 [5] 8 [5] 9 [3 5]})

(residuals-independence y x neighbours-matrix with-replacement)
=>
    {:samples {:morans-i-stat [-0.4150026069686525 -0.5675907000921192 -0.2945316358890812 -0.3529457255608657 -0.2701610954724725],
               :geary-c-stat [2.33029230375536 1.355495681680239 0.9943921584348755 1.4344762237619202 1.2853565231416468]},
     :tests {:morans-i-test [-0.5675907000921192 -0.2701610954724725 -0.38004635279663823 2/5],
             :geary-c-test [0.9943921584348755 2.33029230375536 1.4800025781548083 2/5]}}

```````
(require '[regression-tests.sample-tests :refer :all])

`````````
Use 'moran-i-test' to compute a global spatial autocorrelation index Moran's I.

``````````` example from
```````````    vignettes for R package ape v. 3.4 Moran's I
```````````    https://cran.r-project.org/web/packages/ape/vignettes/MoranI.pdf
```````````    E. Paradis Moran's Autocorrelation Coefficient in Comparative Methods Nov. 29, 2015
```````````    example in 2.1 Phylogenetic Distances, pp. 2-4.

(def values1 [4.09434 3.61092 2.37024 2.02815 -1.46968])
(def neighbour-weights1  {[0 1] 0.505744983336052 [0 2] 0.216747850001166
                          [0 3] 0.171300720162211 [0 4] 0.106206446500571
                          [1 0] 0.505744983336052 [1 2] 0.216747850001166
                          [1 3] 0.171300720162211 [1 4] 0.106206446500571
                          [2 1] 0.304848067656604 [2 0] 0.304848067656604
                          [2 3] 0.240928311535057 [2 4] 0.149375553151735
                          [3 1] 0.276243093922652 [3 2] 0.276243093922652
                          [3 0] 0.276243093922652 [3 4] 0.171270718232044
                          [4 1] 0.25 [4 2] 0.25
                          [4 3] 0.25 [4 0] 0.25})

(moran-i-test values1 neighbour-weights1)
=> -0.07312179438450675

`````````
Use 'geary-c-test' to compute a global spatial autocorrelation index Geary's C.

``````````` example from
```````````   Goodchild, M. F. (1986). Spatial Autocorrelation, CATMOG 47.
```````````                            Geo Books, Norwich, UK, 56 pp.
```````````   example in 1.3.1 Geary's index (area objects, interval attributes), p. 14.

(def values2 [3 2 2 1])
(def neighbour-weights2 {[0 1] 1 [0 2] 1 [0 3] 1 [1 0] 1 [1 3] 1
                         [2 0] 1 [2 3] 1 [3 0] 1 [3 1] 1 [3 2] 1})

(geary-c-test values2 neighbour-weights2)
=> 6/5

`````

## References
  [API Documentation] [doc] folder at the root of this distribution

  [1] Anderson, M. (2001). Permutation tests for univariate or multivariate analysis of variance and regression.
      Canadian Journal of Fisheries and Aquatic Sciences, 58(3): 626-639. DOI: 10.1139/f01-004.
  [2] Freedman, D., & Lane, D. (1983). A Nonstochastic Interpretation of Reported Significance Levels.
      Journal of Business & Economic Statistics, 1(4): 292-298. DOI: 10.2307/1391660.
  [3] Efron, B. (1979). Bootstrap Methods: Another Look at the Jackknife.
      Annals of Statistics, 7(1): 1-26. DOI:10.1214/aos/1176344552.
  [4] Efron, B., & Tibshirani, R. (1993). An Introduction to the Bootstrap.
      New York: Chapman and Hall.
  [5] Geary, R. (1954). The Contiguity Ratio and Statistical Mapping.
      The Incorporated Statistician, 5(3): 115-145. DOI: 10.2307/2986645.
  [6] Moran, P. (1950). Notes on Continuous Stochastic Phenomena.
      Biometrika, 37(1-2): 17-23. DOI: 10.2307/2332142.
  [7] Lin, K.-P., Long, Z.-H., & Ou, B. (2011). The Size and Power of Bootstrap Tests for Spatial Dependence in a Linear Regression Model.
      Computational Economics, 38(2): 153-171. DOI: 10.1007/s10614-010-9224-0.

## License

Licensed under [MIT](http://opensource.org/licenses/MIT)
