(ns regression-tests.sample-tests-test
  (:require [clojure.test :refer :all]
            [regression-tests.sample-tests :refer :all]
            [utilities.floating-point-comparison :refer :all]))

;;; tests

; from
;       vignettes for R package ape v. 3.4 Moran's I
;       https://cran.r-project.org/web/packages/ape/vignettes/MoranI.pdf
;       E. Paradis Moran's Autocorrelation Coefficient in Comparative Methods Nov. 29, 2015
;       example in 2.1 Phylogenetic Distances, pp. 2-4.

(deftest moran-i-test-test
  (testing "Values should be tested for spatial autocorrelation
    using Moran's I statistics."
    (
     ; Act
     let [moran-i
          (moran-i-test [4.09434 3.61092 2.37024 2.02815 -1.46968]
                         {[0 1] 0.505744983336052 [0 2] 0.216747850001166
                          [0 3] 0.171300720162211 [0 4] 0.106206446500571
                          [1 0] 0.505744983336052 [1 2] 0.216747850001166
                          [1 3] 0.171300720162211 [1 4] 0.106206446500571
                          [2 1] 0.304848067656604 [2 0] 0.304848067656604
                          [2 3] 0.240928311535057 [2 4] 0.149375553151735
                          [3 1] 0.276243093922652 [3 2] 0.276243093922652
                          [3 0] 0.276243093922652 [3 4] 0.171270718232044
                          [4 1] 0.25 [4 2] 0.25
                          [4 3] 0.25 [4 0] 0.25})]

       ; Assert
       (is (double= moran-i
                    -0.0731217943845068)))))

; from
;    Goodchild, M. F. (1986). Spatial Autocorrelation, CATMOG 47.
;                             Geo Books, Norwich, UK, 56 pp.
;    example in 1.3.1 Geary's index (area objects, interval attributes), p. 14.

(deftest geary-c-test-test
  (testing "Values should be tested for spatial autocorrelation
    using Geary's I statistics (for sample)."
    (
     ; Act
     let [geary-c
          (geary-c-test [3 2 2 1]
                        {[0 1] 1 [0 2] 1 [0 3] 1
                         [1 0] 1 [1 3] 1
                         [2 0] 1 [2 3] 1
                         [3 0] 1 [3 1] 1 [3 2] 1})]

       ; Assert
       (is (double= geary-c
                    1.2)))))

; from
;    Goodchild, M. F. (1986). Spatial Autocorrelation, CATMOG 47.
;                             Geo Books, Norwich, UK, 56 pp.
;    example in 1.3.1 Geary's index (area objects, interval attributes), p. 14,
;    corrected with n/(n-1) to get population statistics.

(deftest geary-c-test-population-test
  (testing "Values should be tested for spatial autocorrelation
    using Geary's I statistics (for population)."
    (
     ; Act
     let [geary-c
          (geary-c-test-population [3 2 2 1]
                                   {[0 1] 1 [0 2] 1 [0 3] 1
                                    [1 0] 1 [1 3] 1
                                    [2 0] 1 [2 3] 1
                                    [3 0] 1 [3 1] 1 [3 2] 1})]

       ; Assert
       (is (double= geary-c
                    1.6)))))
