(ns utilities.cmd
  (:require [clojure.string :as string]
            [clojure.tools.cli :as cli]))


;;; private methods

;; Defines command line options
(def ^{:private true} cli-options
  [["-t" "--trace" "Print stack trace"
    :id :trace
    :default false]
   ["-h" "--help" "Print command help"]])

(defn- usage
  "Returns program usage description.
    short-description: Short program description.
    options-summary: Description of command line options.
    arguments-summary: Description of command line arguments.
   returns: Program usage description."
  [short-description
   options-summary
   arguments-summary]
  (->> [short-description
        ""
        (->> ((juxt #(->> (:required %)
                          (map first))
                    #(->> (:optional %)
                          (map (fn[[arg _]]
                                 (str "[" arg "]")))))
              arguments-summary)
             (apply concat)
             (string/join \space)
             (str "Usage: program-name [options] "))
        ""
        "Options:"
        options-summary
        ""
        "Arguments:"
        (->> ((juxt :required
                    :optional)
              arguments-summary)
             (apply concat)
             (map (fn[[arg desc]]
                    (str "  " arg \tab desc)))
             (string/join \newline))
        ""
        "Please refer to the manual page for more information."]
       (string/join \newline)))

(defn- exit
  "Terminates program execution.
    status: Exit status.
    message: Termination message."
  [status messsage]
  (do
    (println messsage)
    (System/exit status)))


;;; public methods

(defn terminal
  "Executes program in terminal.

  Arguments: {_ :short-description
              _ :args
              _ :args-description
              _ :execute}

  Run [:execute] with command line arguments in [:args].

  Supplies program with options.
  Supported command line options are -h (--help)
                                     -t (--trace)

  Wraps unhandled exceptions. If -t is set, returns full stack trace.
  Otherwise, returns only exception message.

  Help provides short program description, description of command line
  options, description of command line arguments, etc.

  Short description is supplied with [:short-description] key.

  Description of command line arguments is set in [args-description] key.
  The format for arguments description is
      {:required
       [[argument-name1 argument-description1]
        [argument-name2 argument-description2]
         ...]
       :optional
       [[optional-argument-name1 argument-description1]
        [optional-argument-name2 argument-description2]
         ...]}

  ## Usage

      (require '[utilities.cmd :refer :all])

      (def arguments
           {:short-desc \"Program short description goes here.\"
            :args '(arg1 arg2 ... [optional-arg1] [optional-arg2] ...)
            :args-desc {:required
                        [[\"arg1\" \"Description for arg1 goes here.\"]
                         [\"arg2\" \"Description for arg2 goes here.\"]
                         ...]
                        :optional
                        [[\"optional-arg1\" \"Description for optional-arg1 goes here.\"]
                         [\"optional-arg2\" \"Description for optional-arg2 goes here.\"]
                         ...]}
            :execute (fn[args](println \"Hello, world!\"))})

      (terminal arguments)

  "
  [{program-short-desc :short-desc
   args :args
   cli-args :args-desc
   execute :execute}]
  (let [{:keys [options
                arguments
                errors
                summary]}
        (cli/parse-opts args
                        cli-options)
        usage-method
        #(usage program-short-desc
                summary
                cli-args)]
    (cond
     (:help options) (exit 1 (usage-method))
     (< (count arguments)
        (count (:required cli-args))) (exit 1 (usage-method))
     (> (count arguments)
        (->> ((juxt :required
                    :optional)
              cli-args)
             (map count)
             (apply +))) (exit 1 (usage-method))
     errors (exit 1 (string/join \newline errors))
     :else
     (try
       (execute arguments)
       (catch Exception e (if (:trace options)
                            (throw e)
                            (println (str "Error: " (.getMessage e)))))
       (finally (shutdown-agents))))))
