(ns utilities.writer
  (:require [clojure.java.io :as io]
            [clojure-csv.core :as csv]))


;;; public methods

(defn csv-file
  "Writes data to a csv file.

   Arguments: target-dir
              file-name
              data

   If [target-dir] does not exists, it is created before saving.

   File content in [data] should be represented as a sequence of sequences of strings.
   The data is stored by rows.

   File is saved to [target-dir]-[file-name].
   Returns an absolute path to the [target-dir].

   ## Usage

      (require '[utilities.writer :refer :all])

      (csv-file \"folder\"
                \"file.csv\"
                [[\"header 1\" \"header 2\"]
                 [\"row_1_col_1\" \"row_1_col_2\"]
                 [\"row_2_col_1\" \"row_2_col_2\"]])
  "
  [target-dir
   file-name
   data]
  (let [parent-dir (io/file target-dir)
        file (io/file parent-dir
                      file-name)]
    (do
      (if-not (.exists parent-dir)
        (.mkdir parent-dir))
      (with-open [writer (io/writer file)]
        (spit writer
              (csv/write-csv data)))
      (identity (.getAbsolutePath parent-dir)))))

(defn txt-file
  "Writes data to a text file.

   Arguments: target-dir
              file-name
              lines

   If [target-dir] does not exists, it is created before saving.

   File content in [lines] should be represented as a sequence of strings.

   File is saved to [target-dir]-[file-name].
   Returns an absolute path to the [target-dir].

   ## Usage

      (require '[utilities.writer :refer :all])

      (txt-file \"folder\"
                \"file.asc\"
                [\"line 1\"
                 \"line 2\"
                 \"line 3\"])"
  [target-dir
   file-name
   lines]
  (let [parent-dir (io/file target-dir)
        file (io/file parent-dir
                      file-name)]
    (do
      (if-not (.exists parent-dir)
        (.mkdir parent-dir))
      (with-open [file-writer (io/writer file)]
        (doseq [line lines]
          (.write file-writer line)
          (.newLine file-writer)))
      (identity (.getAbsolutePath parent-dir)))))
